from qtrf.QtCore      import Signal, Slot
from qtrf.numeric_suffix import prefix_map


class PrefixKeysMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.prefix_keys = prefix_map.copy()

    prefix_key_pressed = Signal(int)

    # helpers

    @Slot()
    def keyPressEvent(self, event):
        key_value = event.key()
        try:
            character = chr(key_value)
        except ValueError:
            # not character
            super().keyPressEvent(event)
            return

        if character in self.prefix_keys:
            event.accept()
            self.prefix_key_pressed.emit(key_value)
        elif character.lower() in self.prefix_keys:
            event.accept()
            self.prefix_key_pressed.emit(key_value)
        else:
            # not prefix
            super().keyPressEvent(event)



from .amps.label            import AmpsLabel
from .amps.line_edit        import AmpsLineEdit
from .celsius.label         import CelsiusLabel
from .celsius.line_edit     import CelsiusLineEdit
from .db.label              import dBLabel
from .db.line_edit          import dBLineEdit
from .dbm.label             import dBmLabel
from .dbm.line_edit         import dBmLineEdit
from .file                  import FileWidget
from .hz.label              import HzLabel
from .hz.line_edit          import HzLineEdit
from .int_value.label       import IntValueLabel
from .int_value.line_edit   import IntValueLineEdit
from .ip_address.line_edit  import IPAddressLineEdit
from .meters.label          import MetersLabel
from .meters.line_edit      import MetersLineEdit
from .ohms.label            import OhmsLabel
from .ohms.line_edit        import OhmsLineEdit
from .points.line_edit      import PointsLineEdit
from .ports.label           import PortsLabel
from .ports.line_edit       import PortsLineEdit
from .seconds.line_edit     import SecondsLineEdit
from .touchstone            import TouchstoneWidget
from .user_corrections      import UserCorrectionsWidget
from .float_value.label     import FloatValueLabel
from .float_value.line_edit import FloatValueLineEdit
from .volts.label           import VoltsLabel
from .volts.line_edit       import VoltsLineEdit


# exports
__all__ = [
    "AmpsLabel",
    "AmpsLineEdit",
    "CelsiusLabel",
    "CelsiusLineEdit",
    "dBLabel",
    "dBLineEdit",
    "dBmLabel",
    "dBmLineEdit",
    "FileWidget",
    "HzLabel",
    "HzLineEdit",
    "IntValueLabel",
    "IntValueLineEdit",
    "IPAddressLineEdit",
    "MetersLabel",
    "MetersLineEdit",
    "OhmsLabel",
    "OhmsLineEdit",
    "PointsLineEdit",
    "PortsLabel",
    "PortsLineEdit",
    "SecondsLineEdit",
    "TouchstoneWidget",
    "UserCorrectionsWidget",
    "FloatValueLabel",
    "FloatValueLineEdit",
    "VoltsLabel",
    "VoltsLineEdit",
]



from decimal import Decimal, InvalidOperation, ROUND_HALF_UP



def to_base_and_prefix(value, prefix_map):
    if not prefix_map:
        # no prefixes
        return value, ''

    # do not prefix zero
    if value == 0:
        return 0, ''

    value = Decimal(value)

    # get prefix
    prefix = list(prefix_map)[-1]
    for i in prefix_map:
        if abs(value) < prefix_map[i] * Decimal('1e3'):
            prefix = i
            break

    # calculate base value
    prefix_value = prefix_map[prefix]
    base_value   = value / prefix_value

    return base_value, prefix


def round_decimal_places(value, decimal_places):
    value = Decimal(value)
    if decimal_places >= 1:
        zeros  = '0' * (decimal_places-1)
        format = Decimal('.' + zeros + '1')
    else:
        format = Decimal('1.')
    return value.quantize(format, ROUND_HALF_UP)



from qtrf.QtCore import QEasingCurve, QPropertyAnimation
from time import sleep


# constants
OFFSET_PX    =  10
AMPLITUDE_PX =   3
PERIOD_S     = 300e-3
DURATION_MS  = 500


def shake(window):
    # window starts shifted to the right
    start_position = window.geometry()
    start_position.moveRight(start_position.right() + OFFSET_PX)

    # window stops in original position
    end_position = window.geometry()

    # define window shake "curve"
    shake_curve = QEasingCurve(QEasingCurve.OutElastic)
    shake_curve.setAmplitude(AMPLITUDE_PX)
    shake_curve.setPeriod(PERIOD_S)

    # define animation properties
    animation = QPropertyAnimation(window, b'geometry')
    animation.setStartValue(start_position)
    animation.setEndValue  (end_position)
    animation.setDuration(DURATION_MS)
    animation.setEasingCurve(shake_curve)

    # start
    animation.start()
    return animation



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'file.ui'
##
## Created by: Qt User Interface Compiler version 6.6.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from qtrf.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from qtrf.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from qtrf.QtWidgets import (QApplication, QHBoxLayout, QLineEdit, QPushButton,
    QSizePolicy, QWidget)

class Ui_FileWidget(object):
    def setupUi(self, FileWidget):
        if not FileWidget.objectName():
            FileWidget.setObjectName(u"FileWidget")
        self.horizontalLayout = QHBoxLayout(FileWidget)
#ifndef Q_OS_MAC
        self.horizontalLayout.setSpacing(-1)
#endif
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.filename = QLineEdit(FileWidget)
        self.filename.setObjectName(u"filename")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.filename.sizePolicy().hasHeightForWidth())
        self.filename.setSizePolicy(sizePolicy)
        self.filename.setFocusPolicy(Qt.NoFocus)
        self.filename.setReadOnly(True)

        self.horizontalLayout.addWidget(self.filename)

        self.open = QPushButton(FileWidget)
        self.open.setObjectName(u"open")
        sizePolicy1 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.open.sizePolicy().hasHeightForWidth())
        self.open.setSizePolicy(sizePolicy1)
        self.open.setMinimumSize(QSize(60, 0))
        self.open.setMaximumSize(QSize(60, 16777215))

        self.horizontalLayout.addWidget(self.open)

        self.clear = QPushButton(FileWidget)
        self.clear.setObjectName(u"clear")
        sizePolicy1.setHeightForWidth(self.clear.sizePolicy().hasHeightForWidth())
        self.clear.setSizePolicy(sizePolicy1)
        self.clear.setMinimumSize(QSize(60, 0))
        self.clear.setMaximumSize(QSize(60, 16777215))

        self.horizontalLayout.addWidget(self.clear)

        self.horizontalLayout.setStretch(0, 1)

        self.retranslateUi(FileWidget)

        QMetaObject.connectSlotsByName(FileWidget)
    # setupUi

    def retranslateUi(self, FileWidget):
        self.open.setText(QCoreApplication.translate("FileWidget", u"...", None))
        self.clear.setText(QCoreApplication.translate("FileWidget", u"X", None))
        pass
    # retranslateUi




from .helpers                 import ip_address_validator
from qtrf.QtCore           import Slot
from qtrf.QtWidgets        import QLineEdit
from qtrf.mixins.special_keys import SpecialKeysMixin


class IPAddressLineEdit(SpecialKeysMixin, QLineEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.update_validator()
        self.connect_qt_signals_and_slots()

    @Slot()
    def focusInEvent(self, event):
        super().focusInEvent(event)
        self.selectAll()

    @Slot()
    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        self.selectAll()


    # helpers

    def connect_qt_signals_and_slots(self):
        self.special_key_pressed.connect(self.selectAll)

    def update_validator(self):
        self.setValidator(ip_address_validator())



from .helpers import to_base_and_prefix, round_decimal_places
from decimal  import Decimal, InvalidOperation


# validate str
def regex_str(prefixes, units, include_negative):
    sign_regex  = r'\-?' if include_negative else ''
    value_regex = r'[0-9]+(\.[0-9]*)?'
    if prefixes:
        prefixes_str   = ''.join(prefixes)
        prefixes_regex = f'[{prefixes_str}]?'
    else:
        prefixes_regex = ''
    units_regex = units or ''
    return f'{sign_regex}{value_regex} {prefixes_regex}{units}'.strip()


# str => float if successful else None
def to_decimal(text, units, prefix_map=None):
    # clean input
    try:
        text = text.strip().replace(' ', '')
    except AttributeError:
        return None
    if not text:
        return None

    # chop off units
    if units and text.endswith(units):
        unit_pos = -1 * len(units)
        text     = text[:unit_pos]

    # separate prefix
    prefix = ''
    if prefix_map and text[-1] in prefix_map:
        prefix = text[-1]
        text   = text[:-1]

    # get decimal value
    try:
        _decimal = Decimal(text)
    except InvalidOperation:
        return None

    if not prefix:
        return _decimal

    prefix_value = prefix_map[prefix]
    return _decimal * prefix_value


# number => str
def to_str(value, decimal_places, prefix_map, units):
    base, prefix = to_base_and_prefix(value, prefix_map)
    base = round_decimal_places(base, decimal_places)
    return f'{base} {prefix}{units}'.strip()



from .mixins           import PortsListMixin, ValidatorMixin
from .mixins.ports     import to_list
from qtrf.QtCore    import Slot
from qtrf.QtWidgets import QLineEdit
from qtrf.mixins       import SpecialKeysMixin


class PortsLineEdit(SpecialKeysMixin, ValidatorMixin, PortsListMixin, QLineEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.special_key_pressed.connect(self.enter_value)

    @Slot()
    def focusInEvent(self, event):
        super().focusInEvent(event)
        self.selectAll()

    @Slot()
    def focusOutEvent(self, event):
        event.accept()
        self.enter_value()
        self.deselect()

    @Slot()
    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        self.selectAll()


    # helpers

    def enter_value(self):
        self.list = to_list(self.text())



single_port_regex   = r'[1-9]\d*'
end_port_regex      = f'\\-{single_port_regex}'
port_or_range_regex = f'{single_port_regex}({end_port_regex})?'
ports_list_regex    = f'^{port_or_range_regex}(,\\s*{port_or_range_regex})*$'



from .ui_file                   import Ui_FileWidget
from pathlib                    import Path
from qtrf.QtCore             import Signal, Slot
from qtrf.QtWidgets          import QFileDialog, QWidget


# constants
FILENAME_INDEX = 0


class FileWidget(QWidget):

    # constructor

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.ui = Ui_FileWidget()
        self.ui.setupUi(self)
        self._file_path = None
        self.caption    = None
        self.filter     = ''
        self.save       = False
        self.last_path  = '.'
        self.connect_qt_signals_and_slots()


    # last path (pathlib Path object)

    @property
    def last_path(self):
        return self._last_path


    @last_path.setter
    def last_path(self, last_path):
        self._last_path = Path(last_path)


    # file_path

    @property
    def file_path(self):
        return self._file_path


    @file_path.setter
    def file_path(self, file_path):

        # clear file path?
        if not file_path and self.file_path:
            self._file_path = None
            self.update_view()
            self.file_path_changed.emit(None)
            return

        # change file path?
        if file_path and file_path != self.file_path:
            self._file_path = Path(file_path)
            self._last_path = self.file_path.parent
            self.update_view()
            self.file_path_changed.emit(str(self.file_path))


    # signal: file path changed
    file_path_changed = Signal(str)


    # slot: choose file dialog

    @Slot(str)
    def choose_file(self):
        file_path = self.dialog()
        if not file_path:
            return
        self.file_path = file_path


    # slot: clear file_path

    @Slot()
    def clear(self):
        self.file_path = None


    # helpers


    def default_caption(self):
        return 'Save...' if self.save else 'Open...'


    def dialog(self):
        dialog   = QFileDialog.getSaveFileName if self.save else QFileDialog.getOpenFileName
        settings = {
            "parent":  self.window(),
            "caption": self.default_caption(),
            "dir":     str(self.last_path),
            "filter":  self.filter
        }
        return dialog(*settings.values())[FILENAME_INDEX]


    def update_view(self):
        # clear?
        if not self._file_path:
            self.ui.filename.clear()
            return

        # display file_path
        self.ui.filename.setText(self.file_path.name)


    def connect_qt_signals_and_slots(self):
        self.ui.open .clicked.connect(self.choose_file)
        self.ui.clear.clicked.connect(self.clear)



try:

    # dumpy pyside6
    from PySide6.QtGui import *


except ModuleNotFoundError:

    # pyside6 not installed


    try:

        # dump pyside2
        from PySide2.QtGui import *

        # monkeypatch Qt 6 regex validator name change
        QRegularExpressionValidator = QRegExpValidator

    except ModuleNotFoundError:

        # pyside2 not installed

        # pyside not found
        raise ModuleNotFoundError('qtrf requires either PySide6 or PySide2')



from .regex         import ports_list_regex
from qtrf.QtCore import QRegularExpression
from qtrf.QtGui  import QRegularExpressionValidator


def validator():
    return QRegularExpressionValidator(QRegularExpression(ports_list_regex))



try:

    # dumpy pyside6
    from PySide6.QtWidgets import *


except ModuleNotFoundError:

    # pyside6 not installed


    try:

        # dumpy pyside2
        from PySide2.QtWidgets import *

    except ModuleNotFoundError:

        # pyside2 not installed

        # pyside not found
        raise ModuleNotFoundError('qtrf requires either PySide6 or PySide2')



from qtrf.widgets.float_value.label import FloatValueLabel


class CelsiusLabel(FloatValueLabel):
    def __init__(self, parent=None):
        FloatValueLabel.__init__(self, parent)
        self.prefix_keys    = {}
        self.units          = '°C'
        self.decimal_places = 1



from qtrf.shake     import AnimationsList, shake
from qtrf.QtCore import Slot


# Use requires:
# - class be top-most parent QWidget
# (e.g. window or mainwindow)
class ShakeMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # keep animations out of
        # python garbage collection
        self.animations = AnimationsList()

    Slot()
    def shake(self):
        animation = shake(self)
        self.animations.add(animation)



from qtrf.QtCore import QObject, QPropertyAnimation, Slot
from threading      import RLock


# constants
STOPPED = QPropertyAnimation.Stopped


# filter
def is_stopped(animation):
    return animation.state() == STOPPED


class AnimationsList(QObject):
    def __init__(self):
        super().__init__()

        self.lock       = RLock()
        self.animations = []

    def __del__(self):
        with self.lock:
            self.stop_all()
            self.prune_all()

    def add(self, animation):
        with self.lock:
            self.clear()

            # add
            self.animations.append(animation)

            # prune on finished
            animation.finished.connect(self.prune)

    @Slot()
    def prune(self):
        with self.lock:
            for animation in self.animations:
                if not is_stopped(animation):
                    # animation is still active.
                    # leave it alone
                    continue

                # animation is stopped.
                # need to remove it.
                self.prune_animation(animation)


    # helpers

    def stop_all(self):
        with self.lock:
            for animation in self.animations:
                animation.stop()

    def prune_animation(self, animation):
        with self.lock:
            assert animation in self.animations

            # find index
            index = self.animations.index(animation)
            assert index != -1

            # delete python reference
            del(self.animations[index])

            # mark for Qt garbage collection
            animation.deleteLater()

    def prune_all(self):
        with self.lock:
            for animation in self.animations:
                self.prune_animation(animation)

    def clear(self):
        self.stop_all()
        self.prune_all()



from .int_value          import IntValueMixin
from .helpers            import int_validator
from .special_keys       import SpecialKeysMixin
from qtrf.QtCore      import Slot
from qtrf.helpers        import int_or_none


class IntLineEditMixin(SpecialKeysMixin, IntValueMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.include_negative = True
        self.update_validator()
        self.connect_qt_signals_and_slots()

    @Slot()
    def focusInEvent(self, event):
        super().focusInEvent(event)
        self.selectAll()

    @Slot()
    def focusOutEvent(self, event):
        super().focusOutEvent(event)
        self.enter_value()
        self.deselect()

    @Slot()
    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        self.selectAll()

    def connect_qt_signals_and_slots(self):
        self.special_key_pressed.connect(self.enter_value)


    # helpers

    def enter_value(self):
        self.value = int_or_none(self.text())
        self.selectAll()

    def update_validator(self):
        self.setValidator(int_validator(self.include_negative))



from qtrf.widgets.float_value.line_edit import FloatValueLineEdit


class dBLineEdit(FloatValueLineEdit):
    def __init__(self, parent=None):
        FloatValueLineEdit.__init__(self, parent)
        self.prefix_keys = {}
        self.units       = 'dB'
        self.update_validator()



from qtrf.QtWidgets import QLabel
from qtrf.mixins       import FloatValueMixin


class FloatValueLabel(FloatValueMixin, QLabel):
    def __init__(self, parent=None):
        super().__init__(parent)



from .helpers                        import prefix_map, units
from qtrf.widgets.float_value.label  import FloatValueLabel


class MetersLabel(FloatValueLabel):
    def __init__(self, parent=None):
        FloatValueLabel.__init__(self, parent)
        self.prefix_keys = prefix_map
        self.units      = units



from qtrf.QtCore      import Signal
from qtrf.helpers        import int_or_none


class IntValueMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = None

    # value
    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        value = int_or_none(value)
        if self._value == value:
            # value is unchanged
            self._update_text()
            return

        # set
        self._value = value
        self._update_text()
        self.value_changed.emit(self.value)

    # signals
    value_changed = Signal(int)

    # helpers

    def _update_text(self):
        if self.value is None:
            self.clear()
        else:
            self.setText(str(self.value))



from .float_line_edit import FloatLineEditMixin
from .float_value     import FloatValueMixin
from .int_line_edit   import IntLineEditMixin
from .int_value       import IntValueMixin
from .prefix_keys     import PrefixKeysMixin
from .shake           import ShakeMixin
from .special_keys    import SpecialKeysMixin


# exports
__all__ = [
    'FloatLineEditMixin',
    'FloatValueMixin',
    'IntLineEditMixin',
    'IntValueMixin',
    'PrefixKeysMixin',
    'ShakeMixin',
    'SpecialKeysMixin',
]



from qtrf.QtCore      import QRegularExpression
from qtrf.QtGui       import QRegularExpressionValidator
from qtrf.numeric_suffix import regex_str


POSITIVE_INT_REGEX =   r'(0|[1-9]\d+)'
INT_REGEX          = r'-?(0|[1-9]\d+)'


def float_validator(prefix_map, units, include_negative):
    _regex = regex_str(prefix_map, units, include_negative)
    return QRegularExpressionValidator(QRegularExpression(_regex))


def int_validator(include_negative):
    regex = INT_REGEX if include_negative else POSITIVE_INT_REGEX
    return QRegularExpressionValidator(QRegularExpression(regex))



def to_list(ports_str):
    ports_str = ports_str.strip()

    # empty list?
    if not ports_str:
        return []

    # clean up trailing partial entry
    if ports_str[-1] == '-':
        ports_str = ports_str[:-1]
    if ports_str[-1] == ',':
        ports_str = ports_str[:-1]

    # empty list?
    if not ports_str:
        return []

    ports = list()
    parts = ports_str.split(',')
    for part in parts:
        part = part.strip()
        if '-' not in part:
            # single port
            port = int(part)
            if not port in ports:
                ports.append(port)
        else:
            # range
            start_port, stop_port = [int(i) for i in part.split('-')]
            for port in range(start_port, stop_port + 1):
                if not port in ports:
                    ports.append(port)
    ports.sort()
    return ports

def to_str(ports_list):
    if not ports_list:
        return ''

    # sort
    ports_list = ports_list.copy()
    ports_list.sort()

    parts = []
    while ports_list:
        port1 = ports_list.pop(0)
        if not ports_list:
            # no more ports
            parts.append(str(port1))
            break

        port2 = ports_list[0]
        if port2 != port1 + 1:
            # single port
            parts.append(str(port1))
            continue

        # range

        # take port2
        ports_list.pop(0)

        # find end of range
        increment = 1
        while ports_list and ports_list[0] == port1 + increment + 1:
            port2      = ports_list.pop(0)
            increment += 1

        parts.append(f'{port1}-{port2}')

    return ', '.join(parts)



from qtrf.QtCore import QRegularExpression
from qtrf.QtGui  import QRegularExpressionValidator


POINTS_REGEX = r'[1-9]\d*'


def points_validator():
    return QRegularExpressionValidator(QRegularExpression(POINTS_REGEX))



from qtrf.QtCore      import Signal
from qtrf.helpers        import float_or_none
from qtrf.numeric_suffix import prefix_map, to_str


# Use requires:
# - class be QWidget
# - methods: text() and setText(text)
class FloatValueMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.decimal_places = 3
        self.prefix_keys     = prefix_map.copy()
        self.units          = 'U'
        self._value         = None

    # value
    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        value = float_or_none(value)
        if self._value == value:
            # value is unchanged
            self._update_text()
            return

        # set
        self._value = value
        self._update_text()
        self.value_changed.emit(self.value)

    # signals
    value_changed = Signal(float)

    # helpers

    def _update_text(self):
        if self.value is None:
            self.clear()
            return

        _text = to_str(self.value, self.decimal_places, self.prefix_keys, self.units)
        self.setText(_text)
